function Install-WithChoco {
    [CmdletBinding()]
    param(
        [Parameter(Position = 0, Mandatory)][string] $package,
        [Parameter(Position = 1)][switch] $f,
        [Parameter(Position = 2)][string[]] $args = @(),
        [Parameter(Position = 3)][string] $packageDir = (Get-Location)
    )

    $packageArgs = @()
    $package = Resolve-FullPath $package
    $packageName = (Split-Path -Leaf $package) -replace '(.*?)\..*', '$1'

    if (Test-Path $package) {
        $packageDir = Split-Path -Parent $package
    }

    $packageArgs += $packageName
    $packageArgs += "-s $packageDir"
    $packageArgs += $args

    if ($f) {
        $packageArgs += '-f'
    }

    choco install @packageArgs
}

function Test-Install {
    [CmdletBinding(SupportsShouldProcess = $true)]
    param(
        [Parameter(Position = 0, Mandatory)][string] $package,
        [Parameter(Position = 1)][string] $baseDir = (Get-Location)
    )

    $profilePath = Split-Path -Parent $profile
    $packageExtensions = (Join-Path $profilePath '..\..\Projects\BoxStarter\Chocolatey-Package.extension\extensions\Chocolatey-Package.extension.psm1' -Resolve)
    $devExtensions = (Join-Path $profilePath '..\..\Projects\BoxStarter\Chocolatey-Dev.extension\extensions\Chocolatey-Dev.extension.psm1' -Resolve)

    Clear-Host

    $packagesRoot = Resolve-Path $baseDir
    $package = $package -replace '\.\\(.*?)\\', '$1'
    $packageFound = (Get-ChildItem $packagesRoot "$package.nuspec" -Recurse)
    $found = @{ $true = $packageFound.FullName; $false = 'Nothing'}[$packageFound -ne $null]

    Write-host $packageFound
    Write-Host "Base Directory: $packagesRoot"
    Write-Host "Package: $package. Found: $found."

    if ($packageFound) {
        $packageDir = Split-Path -Parent $packageFound.FullName
        Write-Host "Package Directory: $packageDir"

        $env:ChocoDebug = $true
        $env:ChocolateyPackageName = $package
        $env:ChocolateyPackageTitle = $package
        $env:ChocolateyPackageFolder = $packageDir
        $installScript = Join-Path $env:ChocolateyPackageFolder 'tools\ChocolateyInstall.ps1'

        if ($PSCmdlet.ShouldProcess("$installScript")) {
            Import-Module $packageExtensions -Force
            Import-Module $devExtensions -Force

            & $installScript
        }
    }
}

function Invoke-BoxStarterBuild {
    [CmdletBinding(SupportsShouldProcess = $true)]
    param(
        [Parameter(Position = 0)][string] $package
    )

    Push-Location $(Get-BoxStarterDirectory)

    & .\build-push.ps1 $package

    Pop-Location
}

function Invoke-BoxStarterPrivateBuild {
    [CmdletBinding(SupportsShouldProcess = $true)]
    param(
        [Parameter(Position = 0)][string] $package
    )

    Push-Location $(Get-BoxStarterPrivateDirectory)

    & .\build-push.ps1 $package

    Pop-Location
}

function Find-MissingPackages {
    $boxStarterRepo = Join-Path $PSScriptRoot '..\..\Projects\BoxStarter' -Resolve
    $boxStarterPrivateRepo = Join-Path $PSScriptRoot '..\..\Projects\BoxStarter-Private' -Resolve

    $publishedPackages = Get-PublishedPackages

    # Get all the nuspec files in the BoxStarter and BoxStarter-Private directories
    # and sort the final array
    $packages = (gci $boxStarterRepo -Recurse *.nuspec | Select-Object -ExpandProperty Name) -replace '.nuspec', ''
    $packages += (gci $boxStarterPrivateRepo -Recurse *.nuspec | Select-Object -ExpandProperty Name) -replace '.nuspec', ''
    $packages = $packages | Sort-Object

    Compare-Object $packages $publishedPackages
}

function Get-PublishedPackages {
    [CmdletBinding()]
    param(
        [Parameter(Position = 0)][string] $publishDirectory = (Join-Path $PSScriptRoot '..\..\BoxStarter' -Resolve)
    )

    $specialCases = @('Microsoft-DotNET3*', 'Microsoft-DotNET4*')
    $specialCasesSearchReplace = @(
        'Microsoft-DotNET3.5'
        'Microsoft-DotNET4.5.2'
        'Microsoft-DotNET4.6.2'
        'Microsoft-DotNET4.6.2-DevPack'
    )

    $publishedPackages = @()

    # For all the special cases in the published directyory
    $specialCasesSearchReplace | % {
        if (gci $publishDirectory -Recurse *.nupkg -Include "$_*") {
            $publishedPackages += $_
        }
    }

    # Get all the built packages in the published directyory
    #$publishedPackages += (Get-ChildItem $publishDirectory -Recurse *.nupkg -Exclude $specialCases | Select-Object -ExpandProperty Name) -replace '(.*?)\.\d+\.\d+.*', '$1'
    $publishedPackages += Get-ChildItem $publishDirectory -Recurse *.nupkg -Exclude $specialCases | Select-Object -ExpandProperty Name

    return $publishedPackages
}

function Remove-AllChocoSources {
    choco source -r | % { choco source remove -n $_.Split('|')[0] }
}

function Remove-AllChocoPackages {
    #choco list -lo -r | ? { $_.Split('|')[0] -ne 'chocolatey' } | % { choco uninstall $_.Split('|')[0] -force }

    $packages = choco list -lo -r | % { $_ -split '\|' | Select-Object -first 1 } | ? { $_ -ne 'chocolatey' }
    choco uninstall $packages -f
}

function Install-AllPackages {
    gci *.nupkg | `
        % { $_.Name -replace '(.*?)\.\d+\.\d+.*', '$1' } | `
        ? { -not (choco list $_ -r -lo) } | `
        % { choco install $_ -s . }
}

function Set-ChocolateySource {
    [CmdletBinding()]
    param(
        [switch] $on
    )

    if ($on) {
        choco source enable -n chocolatey
    }
    else {
        choco source disable -n chocolatey
    }
}

function Push-ToContec {
    [CmdletBinding()]
    param(
        [Parameter(Position = 0)][string] $package
    )

    $package = Resolve-PathSafe $package

    choco push $package -s 'http://nuget.gocontec.com/nuget/Contec/' -f
}