function Connect-VM {
    [CmdletBinding(DefaultParameterSetName = 'name')]
    param(
        [Parameter(ParameterSetName = 'name')][Alias('cn')][String[]] $computerName = $env:ComputerName,
        [Parameter(Position = 0, Mandatory, ValueFromPipelineByPropertyName, ValueFromPipeline, ParameterSetName = 'name')][Alias('VMName')][String] $name,
        [Parameter(Position = 1, Mandatory, ValueFromPipelineByPropertyName, ValueFromPipeline, ParameterSetName = 'id')] [Alias('VMId', 'Guid')][Guid] $id,
        [switch] $startVM
    )

    begin {
        Wr
        ite-Verbose "Initializing InstanceCount, InstanceCount = 0"
        $instanceCount = 0
    }

    process {
        try {
            foreach ($computer in $computerName) {
                Write-Verbose "ParameterSetName is '$($PSCmdlet.ParameterSetName)'"

                if ($PSCmdlet.ParameterSetName -eq 'name') {
                    # Get the VM by Id if Name can convert to a guid
                    if ($name -as [guid]) {
                        Write-Verbose "Incoming value can cast to guid"
                        $vm = Get-VM -Id $name -ErrorAction SilentlyContinue
                    }
                    else {
                        $vm = Get-VM -Name $name -ErrorAction SilentlyContinue
                    }
                }
                elseif ($PSCmdlet.ParameterSetName -eq 'id') {
                    $vm = Get-VM -Id $id -ErrorAction SilentlyContinue
                }
                else {
                    $vm = $inputObject
                }

                if ($vm) {
                    Write-Verbose "Executing 'vmconnect.exe $computer $($vm.Name) -G $($vm.Id) -C $InstanceCount'"
                    vmconnect.exe $computer $vm.Name -G $vm.Id -C $instanceCount
                }
                else {
                    Write-Verbose "Cannot find VM '$name'"
                }

                if ($StartVM -and $vm) {
                    if ($vm.State -eq 'off') {
                        Write-Verbose "StartVM was specified and VM state is 'off'. Starting VM '$($vm.Name)'"
                        Start-VM -VM $vm
                    }
                    else {
                        Write-Verbose "Starting VM '$($vm.Name)'. Skipping, VM is not not in 'off' state."
                    }
                }

                $InstanceCount += 1
                Write-Verbose "InstanceCount = $InstanceCount"
            }
        }
        catch {
            Write-Error $_
        }
    }
}

function New-VirtualMachine {
    [CmdletBinding(SupportsShouldProcess = $true)]
    param(
        [Parameter(Position = 0)][string] $name = 'Win10',
        [Parameter(Position = 1)][UInt64] $size = 40GB,
        [Parameter(Position = 2)][Int64] $memory = 4096MB,
        [Parameter(Position = 3)][string] $iso = 'E:\Temp\Windows 10 Enterprise v1703.2017.04.04.iso',
        [Parameter(Position = 5)][string] $path = 'E:\VMs'
    )

    $switch = Get-VMSwitch 'VMSwitch' -ErrorAction SilentlyContinue
    if (!$switch) {
        $switch = New-VMSwitch 'VMSwitch' -NetAdapterName 'Ethernet'
    }

    $internalSwitch = Get-VMSwitch 'VMSwitchInternal' -ErrorAction SilentlyContinue
    if (!$internalSwitch) {
        $internalSwitch = New-VMSwitch 'VMSwitchInternal' -SwitchType Private
    }

    New-VHD -Path "$path\$name.vhdx" -Fixed -SizeBytes $size
    New-VM $name -VHDPath "$path\$name.vhdx" -MemoryStartupBytes $memory
    Add-VMDvdDrive -VMName $name -Path $iso
    Get-VM -Name $name | Get-VMNetworkAdapter | Connect-VMNetworkAdapter -SwitchName 'VMSwitch'

    Start-VM $name
}

function New-HyperVSnapshot {
    [CmdletBinding(SupportsShouldProcess = $true)]
    param(
        [Parameter(Position = 0, Mandatory)][string] $name,
        [Parameter(Position = 1)][string] $snapshot = 'Base'
    )

    Get-VM $name | Checkpoint-VM -SnapshotName $snapshot
}

function Restore-HyperVSnapshot {
    [CmdletBinding(SupportsShouldProcess = $true)]
    param(
        [Parameter(Position = 0, Mandatory)][string] $name,
        [Parameter(Position = 1)][string] $snapshot = 'Base'
    )

    Get-VM $name | Get-VMSnapshot -Name $snapshot | Restore-VMSnapshot -Confirm:$false
    Get-VM $name | Start-VM
}

function Remove-HyperVSnapshot {
    [CmdletBinding(SupportsShouldProcess = $true)]
    param(
        [Parameter(Position = 0, Mandatory)][string] $name,
        [Parameter(Position = 1, Mandatory)][string] $snapshot
    )

    Get-VM $name | Get-VMSnapshot -Name $snapshot | Remove-VMSnapshot -Confirm:$false
}

function HyperV {
    [CmdletBinding()]
    param(
        [Parameter(Position = 0, Mandatory)][string] $enabled
    )

    if ($enabled -match 'on') {
        Enable-WindowsOptionalFeature -FeatureName 'Microsoft-Hyper-V-All' -Online
    }
    else {
        Disable-WindowsOptionalFeature -FeatureName 'Microsoft-Hyper-V-All' -Online
    }
}

function Remove-VirtualMachine {
    [CmdletBinding()]
    param(
        [Parameter(Position = 0)][string] $name
    )

    get-vm $name | stop-vm | remove-vm
}